import logging
import shlex
import abc
from fractions import Fraction
from typing import NewType, List, Dict, Optional, Union, Literal, Callable, Protocol
from typing_extensions import TypedDict, NotRequired

from jinja2 import Template

from shub_workflow.script import JobKey
from shub_workflow.base import WorkFlowManagerProtocol, Outcome


logger = logging.getLogger(__name__)

Resource = NewType("Resource", str)
ResourceAmmount = Union[int, Fraction]
ResourcesDict = NewType("ResourcesDict", Dict[Resource, ResourceAmmount])
TaskId = NewType("TaskId", str)
OnFinishKey = Union[Outcome, Literal["default", "failed"]]
OnFinishTarget = List[Union[Literal["retry"], TaskId]]


class GraphManagerProtocol(WorkFlowManagerProtocol, Protocol):
    @abc.abstractmethod
    def get_task(self, task_id: TaskId) -> "BaseTask":
        ...


class JobGraphDict(TypedDict):
    tags: Optional[List[str]]
    units: Optional[int]
    on_finish: Dict[OnFinishKey, OnFinishTarget]
    wait_for: List[TaskId]
    retries: int
    wait_time: Optional[int]
    project_id: Optional[int]

    command: NotRequired[List[str]]
    init_args: NotRequired[List[str]]
    retry_args: NotRequired[List[str]]

    origin: NotRequired[TaskId]
    index: NotRequired[int]


class BaseTask(abc.ABC):
    def __init__(
        self,
        task_id: TaskId,
        tags: Optional[List[str]] = None,
        units: Optional[int] = None,
        retries: int = 1,
        project_id: Optional[int] = None,
        wait_time: Optional[int] = None,
        on_finish: Optional[Dict[OnFinishKey, OnFinishTarget]] = None,
    ):
        assert task_id != "retry", "Reserved word 'retry' can't be used as task id"
        self.task_id = task_id
        self.tags = tags
        self.units = units
        self.retries = retries
        self.project_id = project_id
        self.wait_time = wait_time
        self.on_finish: Dict[OnFinishKey, OnFinishTarget] = on_finish or {}

        self.__is_locked: bool = False
        self.__next_tasks: List[BaseTask] = []
        self.__wait_for: List[BaseTask] = []
        self.__required_resources: List[ResourcesDict] = []
        self.__start_callback: Callable[[GraphManagerProtocol, bool], None]

        self.__job_ids: List[JobKey] = []

        self.set_start_callback(self._default_start_callback)

    def set_is_locked(self):
        self.__is_locked = True

    @property
    def is_locked(self) -> bool:
        return self.__is_locked

    def append_jobid(self, jobid: JobKey):
        self.__job_ids.append(jobid)

    def get_scheduled_jobs(self) -> List[JobKey]:
        """
        - Returns the task job ids
        """
        return self.__job_ids

    def add_next_task(self, task: "BaseTask"):
        assert not self.__is_locked, "You can't alter a locked job."
        self.__next_tasks.append(task)

    def add_wait_for(self, task: "BaseTask"):
        assert not self.__is_locked, "You can't alter a locked job."
        self.__wait_for.append(task)

    def add_required_resources(self, resources_dict: ResourcesDict):
        assert not self.__is_locked, "You can't alter a locked job."
        self.__required_resources.append(resources_dict)

    def get_next_tasks(self) -> List["BaseTask"]:
        return self.__next_tasks

    def get_required_resources(self, partial: bool = False) -> List[ResourcesDict]:
        """
        If partial is True, return required resources for each splitted job.
        Otherwise return the resouces required for the full task.
        """
        if not partial:
            return self.__required_resources
        required_resources = []
        parallelization = self.get_parallel_jobs()
        for reqset in self.__required_resources:
            reqres: ResourcesDict = ResourcesDict({})
            for resource, req_amount in reqset.items():
                # Split required resource into N parts.  There are two
                # ideas behind this:
                #
                # - if the job in whole requires some resources, each of
                #   its N parts should be using 1/N of that resource
                #
                # - in most common scenario when 1 unit of something is
                #   required, allocating 1/N of it means that when we start
                #   one unit job, we can start another unit job to allocate
                #   2/N, but not a completely different job (as it would
                #   consume (1 + 1/N) of the resource.
                #
                # Use fraction to avoid any floating point quirks.
                reqres[resource] = Fraction(req_amount, parallelization)
            required_resources.append(reqres)
        return required_resources

    def get_wait_for(self) -> List["BaseTask"]:
        return self.__wait_for

    def as_jobgraph_dict(self) -> JobGraphDict:
        jdict: JobGraphDict = {
            "tags": self.tags,
            "units": self.units,
            "on_finish": self.on_finish,
            "wait_for": [t.task_id for t in self.get_wait_for()],
            "retries": self.retries,
            "project_id": self.project_id,
            "wait_time": self.wait_time,
        }
        self.on_finish["default"] = []
        if self.retries > 0:
            self.on_finish["failed"] = ["retry"]

        return jdict

    def set_start_callback(self, func: Callable[[GraphManagerProtocol, bool], None]):
        self.__start_callback = func

    def get_start_callback(self) -> Callable[[GraphManagerProtocol, bool], None]:
        assert self.__start_callback is not None, "Start callback not initialized."
        return self.__start_callback

    def _default_start_callback(self, manager: GraphManagerProtocol, is_retry: bool):
        pass

    @abc.abstractmethod
    def run(self, manager: GraphManagerProtocol, is_retry=False, index: Optional[int] = None) -> Optional[JobKey]:
        ...

    @abc.abstractmethod
    def get_parallel_jobs(self):
        """
        Returns total number of parallel jobs that this task will consist on.
        """


class Task(BaseTask):
    def __init__(
        self,
        task_id,
        command,
        init_args=None,
        retry_args=None,
        tags=None,
        units=None,
        retries=1,
        project_id=None,
        wait_time=None,
        on_finish=None,
    ):
        """
        id - String. identifies the task.
        command - String. script name or jinja2 template string.
        init_args - List of strings. Arguments and options to add to the command.
        retry_args - List of strings. If given and job is retries, use this list of arguments instead the ones
                     specified in init_args.
        tags - List of strings. tags to add to the scheduled job.
        units - Int. units to use for this scheduled job.
        retries - Int. Max number of retries in case job failed.
        project_id - Int. Run task in given project. If not given, just run in the actual project.
        wait_time - Int. Don't run the task before the given number of seconds after job goes to pending status.
        """
        super().__init__(task_id, tags, units, retries, project_id, wait_time, on_finish)
        assert "." not in self.task_id, ". is not allowed in task name."
        self.command = command
        self.init_args = init_args or []
        self.retry_args = retry_args or []
        self.__template = Template(self.command)

    def as_jobgraph_dict(self) -> JobGraphDict:
        jdict = super().as_jobgraph_dict()
        jdict.update({"command": self.get_commands(), "init_args": self.init_args, "retry_args": self.retry_args})
        return jdict

    def get_commands(self) -> List[str]:
        return self.__template.render().splitlines()

    def get_command(self, index: int = 0) -> List[str]:
        return shlex.split(self.get_commands()[index])

    def get_parallel_jobs(self) -> int:
        """
        Returns total number of parallel jobs that this task will consist on.
        """
        return len(self.get_commands())

    def run(
        self, manager: GraphManagerProtocol, is_retry: bool = False, index: Optional[int] = None
    ) -> Optional[JobKey]:
        command = self.get_command(index or 0)
        tags = []
        if self.tags is not None:
            tags.extend(self.tags)
        if index is None:
            jobname = f"{manager.name}/{self.task_id}"
            tags.append(f"TASK_ID={self.task_id}")
        else:
            jobname = f"{manager.name}/{self.task_id}.{index}"
            tags.append(f"TASK_ID={self.task_id}.{index}")
        if is_retry:
            logger.info('Will retry task "%s"', jobname)
        else:
            logger.info('Will start task "%s"', jobname)
        if is_retry:
            retry_args = self.retry_args or self.init_args
            cmd = command + retry_args
        else:
            cmd = command + self.init_args

        jobid = manager.schedule_script(cmd, tags=tags, units=self.units, project_id=self.project_id)
        if jobid is not None:
            logger.info('Scheduled task "%s" (%s)', jobname, jobid)
            self.append_jobid(jobid)
            return jobid
        return None


class SpiderTask(BaseTask):
    """
    A simple spider task.
    """

    def __init__(
        self,
        task_id,
        spider,
        tags=None,
        units=None,
        retries=1,
        wait_time=None,
        on_finish=None,
        job_settings=None,
        **spider_args,
    ):
        super().__init__(task_id, tags, units, retries, None, wait_time, on_finish)
        self.spider = spider
        self.__spider_args = spider_args
        self.__job_settings = job_settings

    def get_spider_args(self):
        spider_args = self.__spider_args
        if self.__job_settings is not None:
            spider_args.update({"job_settings": self.__job_settings})
        return spider_args

    def as_jobgraph_dict(self):
        jdict = super().as_jobgraph_dict()
        jdict.update({"spider": self.spider, "spider_args": self.get_spider_args()})
        return jdict

    def get_parallel_jobs(self):
        return 1

    def run(self, manager: GraphManagerProtocol, is_retry=False, index: Optional[int] = None) -> Optional[JobKey]:
        assert index is None, "Spider Task don't support parallelization."
        jobname = f"{manager.name}/{self.task_id}"
        if is_retry:
            logger.info('Will retry spider "%s"', jobname)
        else:
            logger.info('Will start spider "%s"', jobname)
        tags = []
        if self.tags is not None:
            tags.extend(self.tags)
        tags.append(f"TASK_ID={self.task_id}")
        jobid = manager.schedule_spider(
            self.spider, tags=tags, units=self.units, project_id=self.project_id, **self.get_spider_args()
        )
        if jobid is not None:
            logger.info('Scheduled spider "%s" (%s)', jobname, jobid)
            self.append_jobid(jobid)
            return jobid
        return None



"""
Utility for cloning ScrapyCloud jobs
Features tagging of cloned from/to jobs (both source and destination) and avoids to clone source jobs already cloned.
By default cloned jobs are scheduled in the same project as source job. If --project-id is given, target project
is overriden.
"""
import logging
from shub_workflow.script import BaseScript
from shub_workflow.utils import dash_retry_decorator


_LOG = logging.getLogger(__name__)


def _transform_cmd(job_cmd):
    if isinstance(job_cmd, list):
        return " ".join(["'%s'" % cmd for cmd in job_cmd[1:]])

    return job_cmd


_COPIED_FROM_META = {
    "job_cmd": ("cmd_args", _transform_cmd),
    "units": (None, None),
    "spider_args": ("job_args", None),
    "tags": ("add_tag", None),
    "job_settings": (None, None),
}


class BaseClonner(BaseScript):
    @staticmethod
    def is_cloned(job):
        for tag in BaseScript._get_metadata_key(job.metadata, "tags") or []:
            if tag.startswith("ClonedTo="):
                _LOG.warning(f"Job {job.key} already cloned. Skipped.")
                return True
        return False

    @dash_retry_decorator
    def is_cloned_by_jobkey(self, jobkey):
        job = self.client.get_job(jobkey)
        return self.is_cloned(job)

    def job_params_hook(self, job_params):
        pass

    def clone_job(self, job_key, units=None, extra_tags=None):
        extra_tags = extra_tags or []
        job = self.client.get_job(job_key)

        spider = self._get_metadata_key(job.metadata, "spider")

        job_params = dict()
        for key, (target_key, _) in _COPIED_FROM_META.items():

            if target_key is None:
                target_key = key

            job_params[target_key] = self._get_metadata_key(job.metadata, key)
            add_tag = job_params.setdefault("add_tag", [])
            add_tag = list(filter(lambda x: not x.startswith("ClonedFrom="), add_tag))
            add_tag.append(f"ClonedFrom={job_key}")
            add_tag.extend(extra_tags)
            job_params["add_tag"] = add_tag
            if units is not None:
                job_params["units"] = units

        self.job_params_hook(job_params)

        for key, (target_key, transform) in _COPIED_FROM_META.items():

            target_key = target_key or key

            if transform is None:

                def transform(x):
                    return x

            job_params[target_key] = transform(job_params[target_key])

        project_id, _, _ = job_key.split("/")
        project = self.get_project(self.project_id or project_id)
        new_job = self.schedule_generic(project, spider, **job_params)
        _LOG.info("Cloned %s to %s", job_key, new_job.key)
        jobtags = self._get_metadata_key(job.metadata, "tags")
        jobtags.append(f"ClonedTo={new_job.key}")
        self._update_metadata(job.metadata, {"tags": jobtags})

        return job, new_job

    @dash_retry_decorator
    def schedule_generic(self, project, spider, **job_params):
        return project.jobs.run(spider, **job_params)


class CloneJobScript(BaseClonner):

    flow_id_required = False

    @property
    def description(self):
        return __doc__

    def parse_project_id(self, args):
        project_id = super().parse_project_id(args)
        if project_id:
            return project_id
        if args.key:
            return args.key[0].split("/")[0]
        if args.tag_spider:
            return args.tag_spider.split("/")[0]

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument(
            "--key",
            type=str,
            action="append",
            default=[],
            help="Target job key. Can be given multiple times. All must be in same project.",
        )
        self.argparser.add_argument(
            "--tag-spider",
            help="In format <project_id>/<tag>/<spider name>," "clone given spider from given project id, by tag",
        )
        self.argparser.add_argument("--units", help="Set number of units. Default is the same as cloned job.", type=int)

    def run(self):
        if self.args.key:
            keys = filter(lambda x: not self.is_cloned_by_jobkey(x), self.args.key)
        elif self.args.tag_spider:
            keys = []
            project_id, tag, spider = self.args.tag_spider.split("/")
            for job in self.get_project(project_id).jobs.iter(spider=spider, state=["finished"], has_tag=tag):
                if not self.is_cloned_by_jobkey(job["key"]):
                    keys.append(job["key"])
        else:
            self.argparser.error("You must provide either --key or --tag-spider.")

        for job_key in keys:
            try:
                self.clone_job(job_key, self.args.units, self.args.tag)
            except Exception as e:
                _LOG.error("Could not restart job %s: %s", job_key, e)


if __name__ == "__main__":
    script = CloneJobScript()
    script.run()



import os
import re
import logging
from typing import Generator

from pkg_resources import resource_filename

from google.cloud import storage


_GS_FOLDER_RE = re.compile(r"gs://([-\w]+)/(.*)$")

_LOGGER = logging.getLogger(__name__)


def get_credfile_path(module, resource, check_exists=True):
    credfile = resource_filename(module, resource)
    if not check_exists or os.path.exists(credfile):
        return credfile


def set_credential_file_environ(module, resource, check_exists=True):
    credfile = get_credfile_path(module, resource, check_exists)

    assert credfile, "Google application credentials file does not exist."
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credfile


def upload_file(src_path, dest_path):
    storage_client = storage.Client()
    try:
        bucket_name, destination_blob_name = _GS_FOLDER_RE.match(dest_path).groups()
    except AttributeError:
        raise ValueError(f"Invalid destination {dest_path}")
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_filename(src_path)
    _LOGGER.info(f"File {src_path} uploaded to {dest_path}.")


def list_folder(path: str) -> Generator[str, None, None]:
    storage_client = storage.Client()
    m = _GS_FOLDER_RE.match(path)
    if m:
        bucket_name, blob_prefix = m.groups()
    else:
        raise ValueError(f"Invalid path {path} for GCS.")
    bucket = storage_client.bucket(bucket_name)
    for blob in bucket.list_blobs(prefix=blob_prefix):
        yield f"gs://{bucket_name}/{blob.name}"


def download_file(path: str, dest: str):
    storage_client = storage.Client()
    m = _GS_FOLDER_RE.match(path)
    if m:
        bucket_name, blob_name = m.groups()
    else:
        raise ValueError(f"Invalid path {path} for GCS.")
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    with open(dest, "wb") as w:
        blob.download_to_file(w)
    _LOGGER.info(f"File {path} downloaded to {dest}.")


def rm_file(path: str):
    storage_client = storage.Client()
    m = _GS_FOLDER_RE.match(path)
    if m:
        bucket_name, blob_name = m.groups()
    else:
        raise ValueError(f"Invalid path {path} for GCS.")
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    blob.delete()
    _LOGGER.info(f"Deleted {path}.")



# Automatically created by: shub deploy

from setuptools import setup, find_packages

setup(
    name="shub-workflow",
    version="1.11.7.3",
    description="Workflow manager for Zyte ScrapyCloud tasks.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="BSD",
    url="https://github.com/scrapinghub/shub-workflow",
    maintainer="Scrapinghub",
    packages=find_packages(),
    install_requires=(
        "pyyaml>=3.12",
        "scrapinghub[msgpack]>=2.3.1",
        "jinja2>=2.7.3",
        "sqlitedict==1.6.0",
        "s3fs>=0.4.0",
        "boto3>=1.9.92",
        "google-cloud-storage>=1.38.0",
        "bloom-filter2==2.0.0",
        "collection-scanner",
        "tenacity",
        "typing-extensions",
        "scrapy",
    ),
    scripts=[],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    package_data={"shub_workflow": ["py.typed"]},
)
